//go:build windows

/*
   Copyright 2024 Docker Compose CLI authors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package formatter

// handleCtrlZ is a no-op on Windows as SIGSTOP is not supported
func handleCtrlZ() {
	// Windows doesn't support SIGSTOP/SIGCONT signals
	// Ctrl+Z behavior is handled differently by the Windows terminal
}
